#!/usr/bin/env python

from mapies.datastores import VIIRSDataStore, DADataStore
from mapies.grids import RotatedGrid
from omegaconf import OmegaConf

conf = OmegaConf.load('viirs.yaml')
grid = RotatedGrid(**conf.grid)

# Load the VIIRS data in memory. If this turns out to be too time consuming, we could modify the code to enable lazy-loading
viirs = VIIRSDataStore.from_raw(conf.raw)

# Now we need to apply filters (maybe).
viirs.apply_filters(conf.filters.ungridded, inplace=True)

# For VIIRS I think it makes sense to keep just the raw data on disk (the preprocessing is very quick). But if we wanted to keep already pre-processed files, here we should end the script by a viirs.store(), and start the next script (i.e. what follows below) by a viirs = VIIRSDataStore.load()

# Grid the data:
viirs.aggregate(grid, inplace=True)
viirs.apply_filters(conf.filters.gridded, inplace=True)

# Set uncertainties: this is very specific to the DA, and within the DA, very specific to the assimilation of VIIRS data. So I think the best place is here, in the top-level script ...
merr = conf.uncertainties.model
oerr = ((viirs.data.obs * conf.uncertainties.obs.a + conf.uncertainties.obs.b) ** 2 + conf.uncertainties.obs.c ** 2) ** .5
viirs.data.loc[:, ['err']] = (merr ** 2 + oerr.values ** 2) ** .5

# Convert the data to the DA format
da = viirs.to(DADataStore)
da.store(conf.DA.path)

# This is not the only way to do it. We could directly do viirs.store(conf.DA.path, fmt=mapies.formatters.nc_DA). However, the "DADataStore" could have methods that the "VIIRSDataStore" doesn't have. In that case, converting the object from a VIIRSDataStore to a DADataStore allows leveraging these extra capabilities.
# For instance, we could have the following:
# da = viirs.to(DADataStore)
# da.setup_uncertainties(conf.uncertainties)    ==> replace the err calculations just above
# da.make_plots()                               ==> make some DA-specific plots/diagnostics
# da.store(conf.DA.path)