#!/usr/bin/env python

import multiprocessing
from dataclasses import dataclass
from mapies.generic import GenericDataStore
# from mapies.datatypes import TotalColumn
from pandas import DataFrame, Timestamp, Timedelta, concat
import xarray as xr
from omegaconf import DictConfig
from typing import List, Self, Callable
from pathlib import Path
from functools import lru_cache, partial
from loguru import logger
from numpy import array
from glob import glob
from datetime import datetime
from tqdm import tqdm
from multiprocessing import cpu_count
from multiprocessing.pool import Pool
from tqdm import tqdm
from mapies.debug import timer
from mapies.formatters import nc_DA


def read_file(
        filename : str,
        obsvar: str = 'Aerosol_Optical_Thickness_550_Land_Ocean_Best_Estimate',
        tref: Timestamp = Timestamp(1993, 1, 1)
) -> DataFrame:
    """
    Parse a single VIIRS file
    :param filename: path to the file
    :param obsvar: variable to read (Aerosol_Optical_Thickness_550_Land_Ocean_Best_Estimate
    :param tref: reference time (TAI93, according to the files attributes, shouldn't be changed)
    :return:
    """
    # logger.debug(f'Read data from {filename}')
    ds = xr.open_dataset(filename)
    df = ds[[obsvar, 'Scan_Start_Time', 'Aerosol_Type_Land_Ocean', 'Aerosol_Optical_Thickness_QA_Flag_Ocean', 'Aerosol_Optical_Thickness_QA_Flag_Land']].to_dataframe().dropna(subset=obsvar).reset_index()
    df.loc[:, 'time'] = (tref + df.Scan_Start_Time)
    return df.rename(columns={obsvar: 'obs', 'Longitude': 'lon', 'Latitude': 'lat'})


@timer
def read_obs(
        files: List[str],
        start: Timestamp | None = None,
        end: Timestamp | None = None,
        obsvar: str = 'Aerosol_Optical_Thickness_550_Land_Ocean_Best_Estimate',
        tref: Timestamp = Timestamp(1993, 1, 1),  # TAI93
        processes: int | None = None
) -> DataFrame:
    """
    Read a bunch of files, extract the variable given by the `obsvar` argument, and return the data as a single dataframe.
    The optional `start` and `end` arguments can be used to limit the temporal extent of the data
    """
    if processes is None :
        processes = cpu_count()
    with Pool(processes=processes) as pp:
        dfs = pp.map(partial(read_file, tref=tref, obsvar=obsvar), tqdm(files))
    df = concat(dfs)
    if start is not None:
        df = df.loc[df.time >= start]
    if end is not None:
        df = df.loc[df.time < end]
    return df


@lru_cache
def get_file_list(patterns: str | List[str]) -> List[str]:
    """
    :param patterns: one or several glob patterns (or exact file names) pointing to the location of files on disk.
    :return: the matching list of files

    The function is cached for better performance, as it would typically called many times with the same arguments, and parsing the file system can take time.
    """

    # 1st make sure we have a list of patterns
    if isinstance(patterns, str):
        patterns = [patterns]

    # Then decode them one by one
    files = []
    for pattern in patterns:
        logger.debug(f"getting file list matching {pattern}")
        files.extend(glob(pattern))

    return files


@dataclass
class VIIRSDataStore(GenericDataStore):
    data: DataFrame | xr.Dataset
    datatype: str = 'TotalColumn'
    griddable: bool = True
    instrument: str = 'VIIRS'

    def load(self, path: str | Path, fmt: Callable = None) -> Self:
        raise NotImplementedError

    @classmethod
    def from_raw(cls, conf: DictConfig) -> Self:
        files = get_file_list(conf.files)

        # Find the date and time based on the file names
        times = [f.split('.')[1:3] for f in files]
        times = array([datetime.strptime(d + h, 'A%Y%j%H%M') for (d, h) in times])

        # Filter the files, adding a bit of safety margin
        files = array(files)[(times >= Timestamp(conf.start) - Timedelta(minutes=15)) & (times <= Timestamp(conf.end) + Timedelta(minutes=30))]

        # Import the data into a pandas DataFrame:
        df = read_obs(files, Timestamp(conf.start), Timestamp(conf.end), conf.obsvar, processes=conf.get('ncpus'))

        return cls(df)

    def store(self, dest: str | Path, fmt: Callable = nc_DA):
        fmt(self, dest)