#!/usr/bin/env python

from dataclasses import dataclass
from abc import ABC, abstractmethod, abstractproperty
from pandas import DataFrame
from typing import List, Self, Callable
from pathlib import Path
from omegaconf import DictConfig
import xarray as xr


# class GenericType(ABC):
#     """
#     This is just a placeholder for now ...
#     """
#     @property
#     @abstractmethod
#     def name(self) -> str:
#         ...
#
#     def __eq__(self, other):
#         if isinstance(other, self.__class__):
#             return self.name == other.name
#         else :
#             return self.name == other


class GenericGrid(ABC):
    @abstractmethod
    def aggregate(self, df: DataFrame) -> DataFrame:
        ...


@dataclass(kw_only=True)
class GenericDataStore(ABC):
    data: DataFrame
    griddable: bool
    datatype: str
    instrument: str
    grid: GenericGrid | None = None

    @abstractmethod
    def store(self, fmt: Callable = None) -> Path | List[Path]:
        ...

    @abstractmethod
    def load(cls, path: str | Path, fmt: Callable = None) -> Self:
        ...

    @abstractmethod
    def from_raw(cls, conf: DictConfig) -> Self:
        ...

    def to(self, cls: Self) -> Self:
        return cls(data=self.data, griddable=self.griddable, datatype=self.datatype, grid=self.grid, instrument=self.instrument)

    def aggregate(self, grid: GenericGrid, inplace: bool = False) -> "GenericDataStore":
        if not self.griddable:
            raise RuntimeError("This data cannot be gridded")
        if self.is_gridded:
            if not inplace:
                return self.new()
            return self
        agg = grid.aggregate(self.data)
        if not inplace:
            return self.new(data=agg, grid=grid)
        self.data = agg
        self.grid = grid

    @property
    def is_gridded(self) -> bool:
        return self.grid is not None

    @property
    def apply_filters(self) -> Callable:
        if self.is_gridded:
            return self.apply_filters_gridded
        else :
            return self.apply_filters_non_gridded

    def apply_filters_gridded(self, filters: DictConfig, inplace: bool = False) -> Self | None:
        # Initially I though of using pandas DataFrame for non gridded data and xarray Dataset for gridded ones, which should lead to using different methods for gridding. But I think it's simpler if we can do everything in pandas. I'm keeping the distinct methods just in case, for now ...
        return self.apply_filters_non_gridded(filters, inplace)

    def apply_filters_non_gridded(self, filters: DictConfig, inplace: bool = False) -> Self | None:
        if inplace:
            self.data = self.data.query(filters.query)
        return self.new(self.data.query(filters.query))

    def new(self, data=None, griddable: bool=None, datatype: str=None, grid: GenericGrid | None = None, instrument: str = None) -> Self:
        """
        Return a new instance of the same class as the current object.
        The attributes are identical by default, but can be overwritten
        (if called without any argument obj.new() is equivalent to obj.copy())
        """
        if data is None: data = self.data
        if griddable is None: data = self.griddable
        if datatype is None: data = self.datatype
        if grid is None: grid = self.grid
        if instrument is None: instrument = self.instrument
        return self.__class__(data=data, griddable=griddable, datatype=datatype, grid=grid, instrument=instrument)
