#!/usr/bin/env python

from mapies.generic import GenericGrid
from dataclasses import dataclass
from pandas import DataFrame
import xarray as xr
from pint import Quantity
from numpy.typing import ArrayLike, NDArray
from typing import Tuple, Dict
from numpy import cos, sin, arctan, pi, nan
from functools import partial, wraps
from loguru import logger


def to_degrees(func):
    @wraps(func)
    def inner(*args, **kwargs):
        """
        Convert *all* arguments to Quantity('deg'):
        - if the values are provided without units (array or primitive types), they are assumed to be in degrees already, the units is simply added
        - if the values are provided in another angle unit (e.g. radians), they will be converted to degrees
        - if the values are in a format that is not compatible with pint (e.g. pandas DataFrame) or have a unit that cannot be converted to degrees, an error will be raised
        """
        args = (Quantity(arg, 'deg') for arg in args)
        kwargs = {k: Quantity(v, 'deg') for k, v in kwargs.items()}
        return func(*args, **kwargs)

    return inner


@to_degrees
def geo_to_rot(*, lats: NDArray, lons: NDArray, center_lat: float | Quantity,
               center_lon: float | Quantity) -> Tuple[NDArray, NDArray]:
    x = cos(center_lat) * cos(lats) * cos(lons - center_lon) + sin(center_lat) * sin(lats)
    y = cos(lats) * sin(lons - center_lon)
    z = - sin(center_lat) * cos(lats) * cos(lons - center_lon) + cos(center_lat) * sin(lats)
    rlat = arctan(z / ((x ** 2 + y ** 2) ** .5))
    rlon = arctan(y / x)
    rlon[x < 0] += pi
    return rlon.to('deg').m, rlat.to('deg').m


def coord_to_indices(
        lons: NDArray,
        lats: NDArray,
        west: Quantity | float,
        south: Quantity | float,
        dlon: Quantity | float,
        dlat: Quantity | float
) -> Tuple[NDArray, NDArray]:
    ilats = ((lats - south) / dlat).astype(int)
    ilons = ((lons - west) / dlon).astype(int)
    return ilons, ilats


@dataclass(kw_only=True)
class RotatedGrid(GenericGrid):
    center_lat: float
    center_lon: float
    dlon: float
    dlat: float
    width: float | None = None
    west: float | None = None
    height: float | None = None
    south: float | None = None

    def __post_init__(self):
        # Make sure we have all required options, and no contradicting options
        assert (self.width is not None) ^ (self.west is not None)
        assert (self.height is not None) ^ (self.south is not None)

        # Reconstruct the missing one(s)
        if self.width is not None:
            self.west = self.center_lon - self.width / 2
        if self.height is not None:
            self.south = self.center_lat - self.height / 2
        self.width = self.east - self.west
        self.height = self.north - self.south

        # Make sure that the dlat and dlon are compatible with the domain dimension
        assert (self.width / self.dlon) == int(self.width / self.dlon)
        assert (self.height / self.dlat) == int(self.height / self.dlat)

        # Attach the coordinate functions
        self.geo_to_rot = partial(geo_to_rot, center_lat=self.center_lat, center_lon=self.center_lon)

    def coord_to_indices(self, *, rlons: NDArray = None, rlats: NDArray = None, lons: NDArray = None, lats: NDArray = None) -> Tuple[NDArray, NDArray]:
        """
        Compute the indices corresponding to lon/lat coordinates. The input coordinates can be provided either as geographical coordinates, through the "lons" and "lats" keywords, or as already rotated coordinates, through the "rlon" and "rlat" keyworkds.

        The method returns the coordinate indices as a tuple of (ilons, ilats) arrays.
        """
        assert (rlons is None) ^ (lons is None), logger.error('Exactly one of the "rlons" (rotated lon) or "lons" (geographical lons) arguments must be provided')
        assert (rlats is None) ^ (lats is None), logger.error('Exactly one of the "rlats" (rotated lat) or "lats" (geographical lats) arguments must be provided')
        assert (rlats is None and rlons is None) ^ (lats is None and lons is None), logger.error("Both coordinates (lats and lons) must be provided in the same coordinate system (rotated or geographical)")
        if rlats is None:
            rlons, rlats = self.geo_to_rot(lons, lats)

        ilons, ilats = coord_to_indices(rlons, rlats, self.west, self.south, self.dlon, self.dlat)

        # Convert (back) to float to enable "NaN" values:
        ilats = ilats.astype(float)
        ilons = ilons.astype(float)

        # Set values outside the domain to NaN
        ilats[ilats < 0] = nan
        ilons[ilons < 0] = nan
        ilats[ilats >= self.nlat] = nan
        ilons[ilons >= self.nlon] = nan
        return ilons, ilats

    @property
    def east(self) -> float:
        return -self.west

    @property
    def north(self) -> float:
        return -self.south

    @property
    def nlat(self) -> int:
        return int(self.height / self.dlat)

    @property
    def nlon(self) -> int:
        return int(self.width / self.dlon)

    def aggregate(self, data: DataFrame) -> DataFrame | None:
        """
        Aggregate on the "rotated_lon" and "rotated_lat" variables (rotated coordinates).
        If they are not present, compute them based on the "lon" and "lat" variables.
        """

        # Compute the rotated coordinates if they are missing.
        if 'rotated_lat' not in data:
            data.loc[:, 'rotated_lon'], data.loc[:, 'rotated_lat'] = self.geo_to_rot(lons=data.lon.values, lats=data.lat.values)

        # Compute the lat/lon indices (on the rotated grid)
        ilons, ilats = self.coord_to_indices(rlons=data.rotated_lon.values, rlats=data.rotated_lat.values)
        data.loc[:, 'ilon'] = ilons
        data.loc[:, 'ilat'] = ilats

        # Filter out the data that's outside the region
        data = data.dropna(subset=['ilon', 'ilat'])
        if len(data) == 0:
            return

        # Add a counter for the number of obs
        data.loc[:, ['nobs']] = 1.

        # Aggregate using the ilon/ilat indices.
        # non-numeric data (e.g. strings) are discarded!
        ndata = data.loc[:, ['ilon', 'ilat', 'nobs']].groupby(['ilon', 'ilat']).sum()
        aggdata = data.groupby(['ilon', 'ilat']).mean()
        aggdata.loc[:, 'nobs'] = ndata.nobs

        return aggdata.reset_index()
